
""" defines some nanosurf colors
Copyright Nanosurf AG 2021
License - MIT
"""
class NSFColorHexStr():
    Orange = "EA390D" 
    Soft_Gray = "555555"
    Soft_White = "F0F0F0"

class NSFColorsTuple():
    Orange = (234,57,13)    
    Soft_Gray = (85,85,85)
    Soft_White = (240,240,240)